# Designed May. 2011 by Pierreden
# Permission to use, copy, modify this software for
# any non-commercial purpose is hereby granted

module LiveIvy

### fix earlier install error of 'images' subfolder
def self.tidyup()
	images = File.join(File.dirname(__FILE__), "gui", "images")
	if File.exist?(images)
		Dir.entries(images).each{|f|
			next if f=='.' || f=='..'
			File.delete(File.join(images, f))
		}
		Dir.delete(images)
	end
end
self.tidyup()
###

  VERSION = "0.6..."
  
  [(begin;Sketchup.temp_dir;rescue;'!';end), 
	(begin;ENV['TEMP'].dup.force_encoding('UTF-8');rescue;ENV['TEMP'];end), 
	(begin;ENV['TMP'].dup.force_encoding('UTF-8');rescue;ENV['TMP'];end), 
	(begin;ENV['TMPDIR'].dup.force_encoding('UTF-8');rescue;ENV['TMPDIR'];end), 
	(begin;ENV['USERPROFILE'].dup.force_encoding('UTF-8');rescue;ENV['USERPROFILE'];end),  
	Dir.pwd, 
	'C:/Temp', 
	'/tmp'
	].each{|d|
		if d && File.directory?(d) && File.writable?(d)
			TEMP = File.expand_path(d)
			break
		end
    }
  
  class IvyTool
    ###
	attr_reader :input_point, :follow_cursor, :thumbs, :leaves
	
    # Tool Methods
    def initialize()
	  @leaves=[]
		LiveIvy::IvyTool.load_definitions
		LiveIvy::Materials.ivy_materials
	  @gui     = UI::WebDialog.new("SketchupIvy #{LiveIvy::VERSION}", false, nil, 450, 680, 150, 150, true)
      pathname = File.expand_path(File.dirname(__FILE__))
      pathname = File.join(pathname, 'gui', 'main.html')
      @gui.set_file(pathname)
      register_callbacks()
      @gui.set_on_close { Sketchup.send_action("selectSelectionTool:") }
      @@seed = self
    end

    def getExtents
      bb = Geom::BoundingBox.new
      bb.add @lines if @lines
      return bb
    end

    def activate
      if (RUBY_PLATFORM.downcase =~ /mswin|mingw/) != nil ### PC
		@gui.show{ self.setup_leaves() }
	  else #MAC
		@gui.show_modal{ self.setup_leaves() }
	  end
	  ###
      @input_point = Sketchup::InputPoint.new()
    end

    def resume(view)
      view.invalidate
    end

    def deactivate(view)
      @gui.close
	  view.invalidate
    end

    def register_callbacks()
      @gui.add_action_callback("change_value"){|dialog, params|
		#Params as [id|newvalue]
		#window.location = 'skp:change_value@1|0.8';
		#Parse params
        params = params.split("|")
        #Saves changed settings to the sketchup registry
        Pdn::LiveIvy::Parameters.set_param(params[0], params[1])
		###
      }
      #Asks for the SketchupIvy.setting_array to populate the dialog
      @gui.add_action_callback("get_settings"){|dialog, params|
        js_command = "set_settings(#{Pdn::LiveIvy::Parameters.to_array});"
        @gui.execute_script(js_command)
      }
	  #Leaves
	  @gui.add_action_callback("leaves_changed"){|dialog, params|
		leaves = params.split('|').compact
		@leaves = []
		leaves.each{|e| @leaves << File.basename(e, ".*") }
		###
		leaves_write_params(@leaves)
		###
	  }
	  #
    end
	
	def self.leaves()
		return @leaves
	end
	
	def setup_leaves()
		thumbs = LiveIvy::IvyTool.thumbs()
		@gui.execute_script("document.getElementById('image_select').options.length=0;")
		thumbs.each_with_index{|p, i|
			path = p.gsub(/'/, "\\\\'")
			js = "appender('#{path}', '#{File.basename(path, ".*")}');"
			@gui.execute_script(js)
		}
		### select from memory - if 'name' remembered select 'name'
		leaves = leaves_read_params() ### an array
		@gui.execute_script("sel_leaves('#{leaves}');")
		#p thumbs
		###
	end
	
	### Leaf params
	def leaves_read_params()
		str = Sketchup.read_default("LiveIvy", "leaves", "")
		return str #.split('|')
	end
	###
	def self.read_leaves()
		str = Sketchup.read_default("LiveIvy", "leaves", "")
		return str #.split('|')
	end
	###
	def leaves_write_params(arr=[])
		if arr[0]
			str = arr.join('|')
		else
			str = ""
		end
		Sketchup.write_default("LiveIvy", "leaves", str)
		return str
	end

	###
	
    def onMouseMove(flags, x, y, view)
      @input_point.pick(view, x, y)
      #if flags & ALT_MODIFIER_MASK == ALT_MODIFIER_MASK && @branch_array
      #  fetch_follow_lines
      #end
      view.invalidate
    end

    def onKeyUp(key, repeat, flags, view)
      if key == 73 || key == 105 # == i
        reset_ivy
      end
    end

    def onLButtonUp(flags, x, y, view)
      if ! @branch_array && (flags == 0 || flags == 256)
        if @input_point.face && @input_point.position
		  
          LiveIvy::IvyLeaf.reload_components(@leaves)
          reset_ivy
          @branch_array = []
          @origin       = @input_point.position + @input_point.face.normal.normalize
          LiveIvy::Branch.new(@origin, 0)
          fetch_stem_lines
        end
      end
      if @branch_array
        if flags == 0 || flags == 256
		  grow_stem
          fetch_stem_lines
        elsif flags & CONSTRAIN_MODIFIER_MASK == CONSTRAIN_MODIFIER_MASK  #Shift
		  birth
          reset_ivy
        elsif flags & COPY_MODIFIER_MASK == COPY_MODIFIER_MASK #Ctrl
		  grow_leafs
          fetch_leaf_points
        elsif flags & ALT_MODIFIER_MASK == ALT_MODIFIER_MASK #Alt
		  @follow_cursor = true
          grow_stem
          @follow_cursor = false
          fetch_stem_lines
        end
      end
      view.invalidate
    end

    def reset_ivy
      read_parameters
      @lines        = nil
      @leafs        = nil
      @follow_lines = nil
      @branch_array = nil
      @faces        = Pdn::Entities::all_faces
      @iterations   = 0
      LiveIvy::Node.reset_nodes
    end

    def fetch_stem_lines
      lines = @branch_array.map { |b| b.point_sequens }.flatten
      @lines = lines if lines.length > 0
    end

    def fetch_leaf_points
      leafs = @branch_array.map { |b| b.leafs_array.map { |l| l.start }.flatten }.flatten
      @leafs = leafs if leafs.length > 0
    end

    def fetch_follow_lines
      follows = @branch_array.map { |b| [@input_point.position, t = b.nodes_array.last.position_point, @input_point.position.distance(t)] }
      @follow_info_lines = follows if follows.length > 0
    end

    def draw_follow_lines(view)
      @follow_cursor = nil
      return unless @follow_info_lines
      @follow_info_lines = @follow_info_lines.sort_by { |r| r[2] }
      closest            = @follow_info_lines.first
      mid                = @follow_info_lines[(@follow_info_lines.length-1).floor/2]
      furthest           = @follow_info_lines.last
      distance           = furthest[2]-closest[2]
      @follow_cursor     =[closest[2], distance]

      view.drawing_color = Sketchup::Color.new "Yellow"
      view.draw_lines [mid [0], mid [1]]
      
      view.drawing_color = Sketchup::Color.new "Green"
      view.draw_lines [closest [0], closest [1]]
      view.drawing_color = Sketchup::Color.new "Red"
      view.draw_lines [furthest [0], furthest [1]]
    end

    def draw(view)
      @input_point.draw(view) if @input_point.valid?
      view.drawing_color = "magenta"
      view.draw_lines @lines if @lines
      view.drawing_color = "green"
      view.draw_lines @leaf_lines if @leaf_lines
	  #@leafs=0 if @leafs.class==Array
      view.draw_points @leafs, 19, 5, "green" if @leafs && @leafs[0]
      #draw_follow_lines(view)
    end

    #
    #
    #LiveIvy Methods
    #
    #

    attr_accessor :branch_array, :faces, :node_tests, :nodes_skipped
    @@temp_counter = 0

    def self.seed
      @@seed
    end

    def temp_counter
      @@temp_counter+=1
    end


    def self.branches
      seed.branch_array
    end

    def params
      @parameters
    end

    def read_parameters
      @parameters = Pdn::LiveIvy::Parameters.read_params
    end

    def grow_stem
      read_parameters
      i              = 0
      @@temp_counter = 0
      @node_tests    = 0
      @nodes_skipped = 0
      t1             = Time.now
      20.times do
        for branch in @branch_array
          branch.grow_stem
        end
        i           +=1
        @iterations += 1
        Sketchup::set_status_text("#{@iterations} / Live branches: #{live = @branch_array.select { |b| b.is_alive }.length} / Dead Branches: #{@branch_array.length-live} / Nodes: #{LiveIvy::Node.nodes_count} / #{i*100/20}%")
      end
      t2 = Time.now
      Sketchup::set_status_text("#{@iterations} / Live branches: #{live = @branch_array.select { |b| b.is_alive }.length} / Dead Branches: #{@branch_array.length-live} / Nodes: #{LiveIvy::Node.nodes_count} / #{i*100/20}% / #{(t2-t1)} sec / Tested Nodes: #{@node_tests}(#{@nodes_skipped})")
      #puts @@temp_counter
    end

    def grow_leafs
      read_parameters
      for branch in @branch_array
        branch.grow_leafs
      end
    end

    def birth
      read_parameters
      @ivygroup             = Sketchup.active_model.entities.add_group
      @ivygroup.description = "Made with Sketchup Ivy"
	  @ivygroup.name = "Sketchup Ivy"
      birth_stem
      birth_leafs
    end

    def birth_stem
      bL       = @branch_array.length
      i        = 0
      branches = @branch_array
      lines    = []

      Sketchup.active_model.start_operation "BuildIvy", true
      for branch in branches
        next unless branch.nodes_array.length > 4
        grp             = @ivygroup.entities.add_group
        grp.description = "Ivy Branches"
        points          = branch.nodes_array.map { |b| b.position_point }
        next if points.length < 3
        radius = params[:branch][:radius] /(branch.parents+1)
        detail = [params[:branch][:detail]-branch.parents, 3].max
        points = Pdn::Array.simplifyAngle(points, 0.90)
        #points = Pdn::Array.simplifyDistance(points, params[:branch][:nodesize])
        #points = Pdn::Interpolate::Catmull.dynamic_process(points, 6)
        points = Pdn::Array.simplifyDistance(points, params[:branch][:radius]*3)

        next unless points && points[1].class == Geom::Point3d
        line = grp.entities.add_edges(points)
        lines.concat(line)
        face               = grp.entities.add_ngon(line.first.start, line.first.start.position.vector_to(line.first.end.position), radius, detail)
        face               = grp.entities.add_face(face)
        face.material      = "ivy_stem"
        face.back_material = "ivy_stem"
        face.reverse!
        face.followme(line)
        i += 1
        Sketchup::set_status_text("Generating Ivy: #{(i*100/bL).to_i}%")
      end

      Sketchup.active_model.commit_operation
      Sketchup.active_model.entities.erase_entities lines
    end


    def birth_leafs
      i  = 0
      lL = @branch_array.map { |b| b.leafs_array.map { |l| l }.flatten }.flatten.length

      Sketchup.active_model.start_operation "BuildLeaves", true
      lgrp             = @ivygroup.entities.add_group
      lgrp.description = "Ivy Leaves"

      for branch in @branch_array
        leafs = branch.leafs_array
        for leaf in leafs
		 if leaf && leaf.component
          compo = lgrp.entities.add_instance(leaf.component, Geom::Transformation.new(leaf.start))
          ratio = params[:leaf][:size]/[compo.bounds.width, compo.bounds.depth, compo.bounds.height].max.to_inch
          compo.transform!(leaf.phi_rotation)
          compo.transform!(leaf.theta_rotation)
          compo.transform!(leaf.skew_rotation)
          compo.transform!( Geom::Transformation.scaling( leaf.start, (1.0+rand)*ratio ) )
          Sketchup::set_status_text("Generating Leaves: #{(i*100/lL).to_i}%")
          i+=1
		 end
        end
      end
	  Sketchup::set_status_text("")
      Sketchup.active_model.commit_operation
    end

    def self.load_definitions()
		textures_dir = File.join(File.expand_path(File.dirname(__FILE__)), "Textures")
		model       = Sketchup.active_model
		definitions = model.definitions
		thumbs=[]
		Dir.entries(textures_dir).each{|f|
			next unless File.extname(f).downcase == ".skp"
			next unless f.downcase =~ /^ivyleaf_/
			d = definitions.load(File.join(textures_dir, f))
			thumbs << d.name+".png"
			th = File.join(TEMP, d.name+".png")
			d.save_thumbnail(th) #unless File.exist?(th)
		}
		### sort so 1, 2, 3.. 9, 10, 11..
		codes=[]
		thumbs.each{|e| codes << e.length.to_s+'|'+e }
		@thumbs = []
		codes.sort.each{|e| @thumbs << File.join( TEMP, e.split('|')[1] ) }
		#p @thumbs
		###
    end
	
	def self.thumbs()
		return @thumbs
	end

  end

  module Materials
    def self.present?(name)
      tmp = Sketchup.active_model.materials.current
      begin
        res = Sketchup.active_model.materials.current = name
      rescue Exception
        Sketchup.active_model.materials.current = tmp
        return false
      end
      Sketchup.active_model.materials.current = tmp
      return true
    end

    def self.ivy_materials
      unless Sketchup.active_model.materials["Ivy_stem"] #LiveIvy::Materials::present?("Ivy_stem")
        self.add_stem_to_library
      end
    end

    def self.add_stem_to_library
      material         = Sketchup.active_model.materials.add("ivy_stem")
      pathname         = File.join(File.expand_path(File.dirname(__FILE__)), "Textures", "stem.jpg")
      material.texture = pathname
	  material.color = [100, 60, 40]
    end
  end
end

### menu
unless file_loaded?(__FILE__)
  ###
  Ivy_toolbar = UI::Toolbar.new("Live Ivy Tool")
  cmdivy = UI::Command.new("Live Ivy Tool"){ Sketchup.active_model.select_tool LiveIvy::IvyTool.new() }
  cmdivy.small_icon = File.join(File.expand_path(File.dirname(__FILE__)), "Textures", "LiveIvy_S.png")
  cmdivy.large_icon = File.join(File.expand_path(File.dirname(__FILE__)), "Textures", "LiveIvy_L.png")
  cmdivy.tooltip = "Generates Ivy: branches & leaves"
  Ivy_toolbar.add_item(cmdivy)
  Ivy_toolbar.show if Ivy_toolbar.get_last_state.abs == 1 # TB_VISIBLE/NEVER
  ###
  UI.menu("Draw").add_item(cmdivy)
  ###
  file_loaded(__FILE__)
end
###
